from http import HTTPStatus
from typing import Any, Dict, Optional, Union, cast

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.error import Error
from ...types import UNSET, Response


def _get_kwargs(
    experiment_id: str,
    source_lan: str,
    *,
    target_id: str,
    target_lan: str,
    x_api_token: str,
) -> Dict[str, Any]:
    headers: Dict[str, Any] = {}
    headers["X-Api-Token"] = x_api_token

    params: Dict[str, Any] = {}

    json_target_id: str
    json_target_id = target_id
    params["target_id"] = json_target_id

    params["target_lan"] = target_lan

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    _kwargs: Dict[str, Any] = {
        "method": "post",
        "url": "/experiments/{experiment_id}/vlan/{source_lan}/connect".format(
            experiment_id=experiment_id,
            source_lan=source_lan,
        ),
        "params": params,
    }

    _kwargs["headers"] = headers
    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[Union[Any, Error]]:
    if response.status_code == HTTPStatus.NO_CONTENT:
        response_204 = cast(Any, None)
        return response_204
    if response.status_code == HTTPStatus.BAD_REQUEST:
        response_400 = Error.from_dict(response.json())

        return response_400
    if response.status_code == HTTPStatus.UNAUTHORIZED:
        response_401 = Error.from_dict(response.json())

        return response_401
    if response.status_code == HTTPStatus.FORBIDDEN:
        response_403 = Error.from_dict(response.json())

        return response_403
    if response.status_code == HTTPStatus.NOT_FOUND:
        response_404 = Error.from_dict(response.json())

        return response_404
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[Any, Error]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    experiment_id: str,
    source_lan: str,
    *,
    client: Union[AuthenticatedClient, Client],
    target_id: str,
    target_lan: str,
    x_api_token: str,
) -> Response[Union[Any, Error]]:
    """Connect a shared vlan in an experiment to another experiment's shared vlan.

    Args:
        experiment_id (str):
        source_lan (str):
        target_id (str):
        target_lan (str):
        x_api_token (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, Error]]
    """

    kwargs = _get_kwargs(
        experiment_id=experiment_id,
        source_lan=source_lan,
        target_id=target_id,
        target_lan=target_lan,
        x_api_token=x_api_token,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    experiment_id: str,
    source_lan: str,
    *,
    client: Union[AuthenticatedClient, Client],
    target_id: str,
    target_lan: str,
    x_api_token: str,
) -> Optional[Union[Any, Error]]:
    """Connect a shared vlan in an experiment to another experiment's shared vlan.

    Args:
        experiment_id (str):
        source_lan (str):
        target_id (str):
        target_lan (str):
        x_api_token (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, Error]
    """

    return sync_detailed(
        experiment_id=experiment_id,
        source_lan=source_lan,
        client=client,
        target_id=target_id,
        target_lan=target_lan,
        x_api_token=x_api_token,
    ).parsed


async def asyncio_detailed(
    experiment_id: str,
    source_lan: str,
    *,
    client: Union[AuthenticatedClient, Client],
    target_id: str,
    target_lan: str,
    x_api_token: str,
) -> Response[Union[Any, Error]]:
    """Connect a shared vlan in an experiment to another experiment's shared vlan.

    Args:
        experiment_id (str):
        source_lan (str):
        target_id (str):
        target_lan (str):
        x_api_token (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, Error]]
    """

    kwargs = _get_kwargs(
        experiment_id=experiment_id,
        source_lan=source_lan,
        target_id=target_id,
        target_lan=target_lan,
        x_api_token=x_api_token,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    experiment_id: str,
    source_lan: str,
    *,
    client: Union[AuthenticatedClient, Client],
    target_id: str,
    target_lan: str,
    x_api_token: str,
) -> Optional[Union[Any, Error]]:
    """Connect a shared vlan in an experiment to another experiment's shared vlan.

    Args:
        experiment_id (str):
        source_lan (str):
        target_id (str):
        target_lan (str):
        x_api_token (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, Error]
    """

    return (
        await asyncio_detailed(
            experiment_id=experiment_id,
            source_lan=source_lan,
            client=client,
            target_id=target_id,
            target_lan=target_lan,
            x_api_token=x_api_token,
        )
    ).parsed
