from http import HTTPStatus
from typing import Any, Dict, Optional, Union

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.aggregate_node import AggregateNode
from ...models.error import Error
from ...types import Response


def _get_kwargs(
    experiment_id: str,
    client_id: str,
    *,
    x_api_token: str,
) -> Dict[str, Any]:
    headers: Dict[str, Any] = {}
    headers["X-Api-Token"] = x_api_token

    _kwargs: Dict[str, Any] = {
        "method": "post",
        "url": "/experiments/{experiment_id}/node/{client_id}/reload".format(
            experiment_id=experiment_id,
            client_id=client_id,
        ),
    }

    _kwargs["headers"] = headers
    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[Union[AggregateNode, Error]]:
    if response.status_code == HTTPStatus.OK:
        response_200 = AggregateNode.from_dict(response.json())

        return response_200
    if response.status_code == HTTPStatus.BAD_REQUEST:
        response_400 = Error.from_dict(response.json())

        return response_400
    if response.status_code == HTTPStatus.UNAUTHORIZED:
        response_401 = Error.from_dict(response.json())

        return response_401
    if response.status_code == HTTPStatus.FORBIDDEN:
        response_403 = Error.from_dict(response.json())

        return response_403
    if response.status_code == HTTPStatus.NOT_FOUND:
        response_404 = Error.from_dict(response.json())

        return response_404
    if response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR:
        response_500 = Error.from_dict(response.json())

        return response_500
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[AggregateNode, Error]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    experiment_id: str,
    client_id: str,
    *,
    client: Union[AuthenticatedClient, Client],
    x_api_token: str,
) -> Response[Union[AggregateNode, Error]]:
    """Reload a node in an experiment

    Args:
        experiment_id (str):
        client_id (str):
        x_api_token (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[AggregateNode, Error]]
    """

    kwargs = _get_kwargs(
        experiment_id=experiment_id,
        client_id=client_id,
        x_api_token=x_api_token,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    experiment_id: str,
    client_id: str,
    *,
    client: Union[AuthenticatedClient, Client],
    x_api_token: str,
) -> Optional[Union[AggregateNode, Error]]:
    """Reload a node in an experiment

    Args:
        experiment_id (str):
        client_id (str):
        x_api_token (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[AggregateNode, Error]
    """

    return sync_detailed(
        experiment_id=experiment_id,
        client_id=client_id,
        client=client,
        x_api_token=x_api_token,
    ).parsed


async def asyncio_detailed(
    experiment_id: str,
    client_id: str,
    *,
    client: Union[AuthenticatedClient, Client],
    x_api_token: str,
) -> Response[Union[AggregateNode, Error]]:
    """Reload a node in an experiment

    Args:
        experiment_id (str):
        client_id (str):
        x_api_token (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[AggregateNode, Error]]
    """

    kwargs = _get_kwargs(
        experiment_id=experiment_id,
        client_id=client_id,
        x_api_token=x_api_token,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    experiment_id: str,
    client_id: str,
    *,
    client: Union[AuthenticatedClient, Client],
    x_api_token: str,
) -> Optional[Union[AggregateNode, Error]]:
    """Reload a node in an experiment

    Args:
        experiment_id (str):
        client_id (str):
        x_api_token (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[AggregateNode, Error]
    """

    return (
        await asyncio_detailed(
            experiment_id=experiment_id,
            client_id=client_id,
            client=client,
            x_api_token=x_api_token,
        )
    ).parsed
